/***********************************************************

Source :  FolderBrowseSample.prg
System :  Windows
Author :  Gnter Beyes
Created:  11.01.2005

Copyright (c) 2005 Gnter Beyes and Xsys Datenkommunikation GmbH.

This source is released under an Apache style licence.

--------
Purpose:
--------

   Demonstrate the BrowseForFolder API in Xbase++.
   Does not require bap.dll.

--------
History:
--------
   11.01.2005 GB - Created.

-------------
Dependencies:
-------------
   Xbase++ 1.80.284 and later ( for Os() function )

***********************************************************/


// BrowseForFolder flags; may be combined
// Ver. 4.71 etc. indicate the minimum required version of shell32.dll.
// --------------------------------------------------------------------

#define BIF_RETURNONLYFSDIRS   0x0001  // For finding a folder to start document searching
#define BIF_DONTGOBELOWDOMAIN  0x0002  // For starting the Find Computer
#define BIF_STATUSTEXT         0x0004   // Top of the dialog has 2 lines of text for BROWSEINFO.lpszTitle
                                        // and one line if this flag is set.
                                        // Passing the message BFFM_SETSTATUSTEXTA to the hwnd
                                        // can set the rest of the text.
                                        // This is not used with BIF_USENEWUI (BIF_NEWDIALOGSTYLE)
                                        // and BROWSEINFO.lpszTitle gets all three lines of text.
#define BIF_RETURNFSANCESTORS  0x0008
#define BIF_EDITBOX            0x0010   // Ver. 4.71 Add an editbox to the dialog
#define BIF_VALIDATE           0x0020   // Ver. 4.71 insist on valid result (or CANCEL)

#define BIF_NEWDIALOGSTYLE     0x0040   // Ver. 5.0  Use the new dialog layout with the ability to resize
                                        // Caller needs to call OleInitialize() before using this API

#define BIF_USENEWUI           (BIF_NEWDIALOGSTYLE + BIF_EDITBOX) // Ver. 5.0

#define BIF_BROWSEINCLUDEURLS  0x0080   // Ver. 5.0 Allow URLs to be displayed or entered.
                                        // (Requires BIF_USENEWUI)

#define BIF_UAHINT             0x0100   // Ver. 6.0
                                        // Add a UA (User Agent) hint to the dialog,
                                        // in place of the edit box.
                                        // May not be combined with BIF_EDITBOX

#define BIF_NONEWFOLDERBUTTON  0x0200   // Ver. 6.0
                                        // Do not add the "New Folder" button to the dialog.
                                        // Only applicable with BIF_NEWDIALOGSTYLE.

#define BIF_NOTRANSLATETARGETS 0x0400   // Ver. 6.0 don't traverse target as shortcut

#define BIF_BROWSEFORCOMPUTER  0x1000  // Browsing for Computers.
#define BIF_BROWSEFORPRINTER   0x2000  // Browsing for Printers
#define BIF_BROWSEINCLUDEFILES 0x4000  // Ver. 4.71 Browsing for Everything
#define BIF_SHAREABLE          0x8000  // Ver. 5.0  shareable resources displayed

// special folder IDs for SHGetSpecialFolderLocation
// -------------------------------------------------

#define CSIDL_DESKTOP                   0x0000  // <desktop>
#define CSIDL_INTERNET                  0x0001  // Internet Explorer (icon on desktop)
#define CSIDL_PROGRAMS                  0x0002  // Start Menu\Programs
#define CSIDL_CONTROLS                  0x0003  // My Computer\Control Panel
#define CSIDL_PRINTERS                  0x0004  // My Computer\Printers
#define CSIDL_PERSONAL                  0x0005  // My Documents
#define CSIDL_FAVORITES                 0x0006  // <user name>\Favorites
#define CSIDL_STARTUP                   0x0007  // Start Menu\Programs\Startup
#define CSIDL_RECENT                    0x0008  // <user name>\Recent
#define CSIDL_SENDTO                    0x0009  // <user name>\SendTo
#define CSIDL_BITBUCKET                 0x000a  // <desktop>\Recycle Bin
#define CSIDL_STARTMENU                 0x000b  // <user name>\Start Menu
#define CSIDL_MYDOCUMENTS               0x000c  // logical "My Documents" desktop icon
#define CSIDL_MYMUSIC                   0x000d  // "My Music" folder
#define CSIDL_MYVIDEO                   0x000e  // "My Videos" folder
#define CSIDL_DESKTOPDIRECTORY          0x0010  // <user name>\Desktop
#define CSIDL_DRIVES                    0x0011  // My Computer
#define CSIDL_NETWORK                   0x0012  // Network Neighborhood (My Network Places)
#define CSIDL_NETHOOD                   0x0013  // <user name>\nethood
#define CSIDL_FONTS                     0x0014  // windows\fonts
#define CSIDL_TEMPLATES                 0x0015
#define CSIDL_COMMON_STARTMENU          0x0016  // All Users\Start Menu
#define CSIDL_COMMON_PROGRAMS           0X0017  // All Users\Start Menu\Programs
#define CSIDL_COMMON_STARTUP            0x0018  // All Users\Startup
#define CSIDL_COMMON_DESKTOPDIRECTORY   0x0019  // All Users\Desktop
#define CSIDL_APPDATA                   0x001a  // Ver. 4.71 <user name>\Application Data
#define CSIDL_PRINTHOOD                 0x001b  // <user name>\PrintHood
#define CSIDL_LOCAL_APPDATA             0x001c  // Ver. 5.0. <user name>\Local Settings\Application Data (non roaming)
#define CSIDL_ALTSTARTUP                0x001d  // non localized startup
#define CSIDL_COMMON_ALTSTARTUP         0x001e  // non localized common startup
#define CSIDL_COMMON_FAVORITES          0x001f
#define CSIDL_INTERNET_CACHE            0x0020
#define CSIDL_COOKIES                   0x0021
#define CSIDL_HISTORY                   0x0022
#define CSIDL_COMMON_APPDATA            0x0023  // Ver. 5.0. All Users\Application Data
#define CSIDL_WINDOWS                   0x0024  // Ver. 5.0. GetWindowsDirectory()
#define CSIDL_SYSTEM                    0x0025  // Ver. 5.0. GetSystemDirectory()
#define CSIDL_PROGRAM_FILES             0x0026  // Ver. 5.0. C:\Program Files
#define CSIDL_MYPICTURES                0x0027  // Ver. 5.0. C:\Program Files\My Pictures
#define CSIDL_PROFILE                   0x0028  // Ver. 5.0. USERPROFILE
//#define CSIDL_SYSTEMX86                 0x0029  // x86 system directory on RISC
//#define CSIDL_PROGRAM_FILESX86          0x002a  // x86 C:\Program Files on RISC
#define CSIDL_PROGRAM_FILES_COMMON      0x002b  // Ver. 5.0. C:\Program Files\Common
// #define CSIDL_PROGRAM_FILES_COMMONX86   0x002c  // x86 Program Files\Common on RISC
#define CSIDL_COMMON_TEMPLATES          0x002d  // All Users\Templates
#define CSIDL_COMMON_DOCUMENTS          0x002e  // All Users\Documents
#define CSIDL_COMMON_ADMINTOOLS         0x002f  // Ver. 5.0. All Users\Start Menu\Programs\Administrative Tools
#define CSIDL_ADMINTOOLS                0x0030  // Ver. 5.0. <user name>\Start Menu\Programs\Administrative Tools
#define CSIDL_CONNECTIONS               0x0031  // Network and Dial-up Connections
#define CSIDL_COMMON_MUSIC              0x0035  // All Users\My Music
#define CSIDL_COMMON_PICTURES           0x0036  // All Users\My Pictures
#define CSIDL_COMMON_VIDEO              0x0037  // All Users\My Video
#define CSIDL_RESOURCES                 0x0038  // %windir%\Resources\, For theme and other windows resources.
#define CSIDL_RESOURCES_LOCALIZED       0x0039  // %windir%\Resources\<LangID>, for theme and other windows specific resources.
#define CSIDL_COMMON_OEM_LINKS          0x003a  // Links to All Users OEM specific apps
#define CSIDL_CDBURN_AREA               0x003b  // USERPROFILE\Local Settings\Application Data\Microsoft\CD Burning
// unused                               0x003c
#define CSIDL_COMPUTERSNEARME           0x003d  // Computers Near Me (from Workgroup membership)

#define CSIDL_FLAG_CREATE               0x8000  // Ver. 5.0. combine with CSIDL_ value to force folder creation in SHGetFolder      Path()
#define CSIDL_FLAG_DONT_VERIFY          0x4000  // combine with CSIDL_ value to return an unverified folder path
#define CSIDL_FLAG_NO_ALIAS             0x1000  // combine with CSIDL_ value to insure non-alias versions of the pidl
#define CSIDL_FLAG_PER_USER_INIT        0x0800  // combine with CSIDL_ value to indicate per-user init (eg. upgrade)
#define CSIDL_FLAG_MASK                 0xFF00  // mask for all possible flag values

// ----------------------------------------------

#INCLUDE "common.CH"
#INCLUDE "dll.CH"
#INCLUDE "os.CH"

// some API-related defines

#DEFINE MAX_PATH        260
#DEFINE SUCCEEDED( hr ) ( ! hr[32] )    // (most-significant bit is 0)

#define GMEM_FIXED      0x0000
#define GMEM_ZEROINIT   0x0040
#define GPTR            (GMEM_FIXED + GMEM_ZEROINIT)

#define CP_ACP          0               // default ANSI code page

PROC Main()

LOCAL cFolder
LOCAL nFlags
LOCAL rootItem

IF SUCCEEDED( CoInitialize(0) )
   nFlags   := (BIF_NEWDIALOGSTYLE + BIF_NOTRANSLATETARGETS)
   rootItem := 'C:\WinDDK'
//?   cFolder := BrowseForFolder( /*nFlags*/, /*rootItem*/, "BrowseForFolder Test" )
      cFolder := BrowseForFolder( nFlags, rootItem , "BrowseForFolder Test" )

   CoUninitialize()
   MsgBox( IIF( empty(cFolder), "(none)", cFolder ), "Folder selected:" )
ENDIF

RETURN

PROC dbesys ; RETURN
PROC appsys ; RETURN

// -------------------------------------------------------

FUNCTION BrowseForFolder( nFlags, xStartFolder, cCaption )

LOCAL rc
LOCAL pidlRoot     := 0
LOCAL pDisplayName := 0
LOCAL pCaption     := 0
LOCAL lpfn         := 0
LOCAL lParam       := 0
LOCAL hwndOwner    := 0
LOCAL nPos
LOCAL BrowseInfo
LOCAL BrowseResult
LOCAL nImageIndex
LOCAL cFolder            := ""
LOCAL cFolderDisplayName := ""

IF setappwindow() <> appdesktop()
   hwndOwner := setappwindow():getHWND()
ENDIF

DEFAULT nFlags TO 0

IF ISNUMBER( xStartFolder )
   rc := SHGetSpecialFolderLocation( hwndOwner, xStartFolder, @pidlRoot )
   IF rc <> 0
      pidlRoot := 0
   ENDIF
ELSEIF ISCHARACTER( xStartFolder )
   pidlRoot := SHSimpleIDListFromPath( xStartFolder )
ENDIF

pDisplayName := CreatePSZ( MAX_PATH )   // create string pointer
IF ISCHARACTER( cCaption )
   pCaption  := CreatePSZ( cCaption )   // create string pointer
ENDIF

// This is the BrowseInfo structure.

                                          // Offset
BrowseInfo := L2bin( hwndOwner )     + ;  //  1  [IN]  handle of owner window
              L2bin( pidlRoot )      + ;  //  5  [IN]  pointer to root item
              L2bin( pDisplayName )  + ;  //  9  [OUT] pointer to selected display name
              L2bin( pCaption )      + ;  // 13  [IN]  pointer to (treeview) caption string
              U2bin( nFlags )        + ;  // 17  [IN]  flags
              L2bin( lpfn )          + ;  // 21  [IN]  pointer to callback function
              L2bin( lParam )        + ;  // 25  [IN]  Parameter to pass to callback function
              L2bin( 0 )                  // 29  [OUT] image index

BrowseResult := SHBrowseForFolderA( @BrowseInfo )

IF BrowseResult <> 0
   // BrowseResult is a pointer to an Item Identifier List structure
   // representing the selected folder's full pathname.

   // Try to obtain folder's full pathname;
   // This will work for file-system items only.
   // For other items (virtual folders), try to obtain the display name instead.

   cFolder := space( MAX_PATH )
   rc := SHGetPathFromIDList( BrowseResult, @cFolder )  // for file-system items
   IF rc = 0
      rc := ILGetDisplayName( BrowseResult, @cFolder )  // for virtual folders
   ENDIF

   IF ( nPos := at( chr(0), cFolder ) ) > 0
      cFolder := left( cFolder, nPos-1 )
   ENDIF

   // get the folder's display name from pointer
   cFolderDisplayName := StringFromPSZ( pDisplayName )

   // get index to the folder's image within system image list
   nImageIndex        := Bin2L( substr( BrowseInfo, 29, 4 ) )

ENDIF // BrowseResult <> 0

// release any allocated memory

IF BrowseResult <> 0
   CoTaskMemFree( BrowseResult )
ENDIF

IF pidlRoot <> 0
   CoTaskMemFree( pidlRoot )
ENDIF

IF pCaption <> 0
   MemFree( pCaption )
ENDIF

IF pDisplayName <> 0
   MemFree( pDisplayName )
ENDIF

RETURN cFolder

// -------------------------------------------
// Some required DLL calls

// Allocate memory for a zero-terminated string;
// optionally copy an Xbase++ string to that memory block.

FUNCTION CreatePSZ( ncString )

LOCAL nSize
LOCAL nPtr := 0

IF ISCHARACTER( ncString )
   nSize := len( ncString )
   IF nSize > 0
      nPtr  := MemAlloc( nSize+1 )    // Plus 1 for terminating zero

      MemCopy( nPtr, ncString, nSize )
   ENDIF
ELSEIF ISNUMBER( ncString ) .AND. ncString <> 0
   // ncString represents required size of memory block
   nPtr  := MemAlloc( ncString+1 )    // Plus 1 for terminating zero
ENDIF

RETURN nPtr

// ------------------------------------------------------------
// allocate a memory block of size <nSize>

FUNCTION MemAlloc( nSize )

LOCAL nPtr := 0

STATIC tpl
IF tpl == NIL
   tpl := DllPrepareCall( "kernel32.dll", DLL_STDCALL, "GlobalAlloc" )
ENDIF

IF nSize <> 0
   nPtr := DllExecuteCall( tpl, GPTR, nSize )
ENDIF

RETURN nPtr

// ------------------------------------------------------------
// release allocated memory

FUNCTION MemFree( ptr )

STATIC tpl
IF tpl == NIL
   tpl := DllPrepareCall( "kernel32.dll", DLL_STDCALL, "GlobalFree" )
ENDIF

RETURN DllExecuteCall( tpl, ptr )

// ------------------------------------------------------------
// copy memory,  from pointer to pointer, or from pointer to Xbase++ string variable

FUNCTION MemCopy( pDest, pSource, nLength )

LOCAL rc

STATIC tpl
IF tpl == NIL
   tpl := DllPrepareCall( "kernel32.dll", DLL_STDCALL, "RtlMoveMemory" )
ENDIF

IF ISCHARACTER( pDest )
   DEFAULT nLength TO len( pDest )
   rc := DllExecuteCall( tpl, @pDest, pSource, nLength )
ELSE
   rc := DllExecuteCall( tpl, pDest, pSource, nLength )
ENDIF

RETURN rc

// ------------------------------------------------------------
// Dereference a string pointer.
// (copy memory block contents to Xbase++ string variable)
// ANSI version

FUNCTION StringFromPSZ( nPtr, nSize )

LOCAL cString := ""

IF nPtr <> 0

   DEFAULT nSize TO StringSizeA( nPtr )  // for Unicode, use ( 2 * StringSizeW( nPtr ) )

   cString := space( nSize )

   MemCopy( @cString, nPtr, nSize )
ENDIF

RETURN cString

// ------------------------------------------------------------
// assuming that lpString points to an Ansi string,
// determine its length, in characters.

FUNCTION StringSizeA( lpString )

STATIC tpl
IF tpl = NIL
   tpl := DllPrepareCall( "kernel32.dll", DLL_STDCALL, "lstrlenA" )
ENDIF

RETURN DllExecuteCall( tpl, lpString )

// ------------------------------------------------------------
// assuming that lpString points to a Unicode string,
// determine its length, in characters.

FUNCTION StringSizeW( lpString )

STATIC tpl
IF tpl = NIL
   tpl := DllPrepareCall( "kernel32.dll", DLL_STDCALL, "lstrlenW" )
ENDIF

RETURN DllExecuteCall( tpl, lpString )

// ------------------------------------------------------------
// release memory that was allocated by the COM subsystem.

PROCEDURE CoTaskMemFree( pv )

STATIC tpl

IF tpl = NIL
   tpl := DllPrepareCall( "ole32.dll", DLL_STDCALL, "CoTaskMemFree" )
ENDIF

DllExecuteCall( tpl, pv )

RETURN

// --------------------------------------------------------------
// Browse For Folder (ANSI version)
// BrowseInfo must be passed by reference

FUNCTION SHBrowseForFolderA( BrowseInfo )

STATIC tpl

IF tpl = NIL
   tpl := DllPrepareCall( "shell32.dll", DLL_STDCALL, "SHBrowseForFolderA" )
ENDIF

RETURN DllExecuteCall( tpl, @BrowseInfo )

// --------------------------------------------------------------
// obtain path string from a pointer to an Item Identifier List (pidl)
// cPath must be passed by reference.

FUNCTION SHGetPathFromIDList( pidl, cPath )

LOCAL rc
LOCAL nPos

STATIC tpl

IF tpl = NIL
   tpl := DllPrepareCall( "shell32.dll", DLL_STDCALL, "SHGetPathFromIDListA" )
ENDIF

cPath := space( MAX_PATH+1 )

rc := DllExecuteCall( tpl, pidl, @cPath )

IF rc <> 0
   IF ( nPos := at( chr(0), cPath ) ) > 0
      cPath := left( cPath, nPos-1 )
   ENDIF
ELSE
   cPath := ""
ENDIF

RETURN rc

// ------------------------------------------------------------------------
// Get pointer to an Item Identifier List (pidl) corresponding to the
// specified Special Folder.
// pidl must be passed by reference.

FUNCTION SHGetSpecialFolderLocation( hwndOwner, nFolder, pidl )

STATIC tpl

IF tpl = NIL
   tpl := DllPrepareCall( "shell32.dll", DLL_STDCALL, "SHGetSpecialFolderLocation" )
ENDIF

pidl := 0

RETURN DllExecuteCall( tpl, hwndOwner, nFolder, @pidl )

// ----------------------------------------------------------
// Undocumented function #162 in shell32.dll
// In WinNT and later, the function is implemented for Unicode only.
// Returns a "pointer to an item identifier list" (PIDL)

FUNCTION SHSimpleIDListFromPath( cPath )

STATIC tpl

IF tpl = NIL
//   tpl := DllPrepareCall( "shell32.dll", DLL_STDCALL, 162 )  'SHSimpleIDListFromPath'
     tpl := DllPrepareCall( "shell32.dll", DLL_STDCALL, 'SHSimpleIDListFromPath' )
ENDIF

IF Os( OS_FAMILY ) == "WINNT"
   cPath := Ansi2Unicode( cPath ) + chr(0)+chr(0)
ENDIF

RETURN DllExecuteCall( tpl, cPath )

// ----------------------------------------------------------
// Undocumented function #15 in shell32.dll.
// In WinNT and later, the function is implemented for Unicode only.
// Returns, in <cBuffer>, the "display name" corresponding to a
// "pointer to an item identifier list" (PIDL).
// cBuffer must be passed by reference.

FUNCTION ILGetDisplayName( pidl, cBuffer )

LOCAL nPos, rc

STATIC tpl

IF tpl = NIL
   tpl := DllPrepareCall( "shell32.dll", DLL_STDCALL, 'SHGetPathFromIDList' )
ENDIF

cBuffer := space( 2*(MAX_PATH+1)  )

rc :=  DllExecuteCall( tpl, pidl, @cBuffer )

IF rc <> 0
   IF Os( OS_FAMILY ) == "WINNT"
      cBuffer := Unicode2Ansi( cBuffer )
   ENDIF
   IF ( nPos := at( chr(0), cBuffer ) ) > 0
      cBuffer := left( cBuffer, nPos-1 )
   ENDIF
ELSE
   cBuffer := ""
ENDIF

RETURN rc

// -----------------------------------------------

FUNCTION Ansi2Unicode( cAnsiString )

LOCAL nSize := 2 * ( len( cAnsiString )+1 )
LOCAL cUnicodeString := replicate( chr(0), nSize )
LOCAL rc

rc := MultiByteToWideChar( CP_ACP, 0, cAnsiString, -1, @cUnicodeString, nSize )

RETURN cUnicodeString

// ---------------------------------------------------------

FUNCTION Unicode2Ansi( cUnicodeString )

LOCAL nLen
LOCAL cAnsiString
LOCAL rc
LOCAL nPos

cUnicodeString += chr(0)+chr(0)      // just to be safe

nLen        := len( cUnicodeString ) // more than required, but so what
cAnsiString := replicate( chr(0), nLen+1 )

rc := WideCharToMultiByte( CP_ACP,;
                           0     ,;
                           cUnicodeString  ,;
                           -1    ,;
                           @cAnsiString ,;
                           nLen  ,;
                           0     ,;
                           0 )

nPos := at( chr(0), cAnsiString )
IF nPos > 0
   cAnsiString := left( cAnsiString, nPos-1 )
ENDIF

RETURN cAnsiString

// -----------------------------------------------

DLLFUNCTION CoInitialize( nReserved ) USING STDCALL FROM ole32.dll

DLLFUNCTION CoUninitialize() USING STDCALL FROM ole32.dll

DLLFUNCTION MultiByteToWideChar( nCodePage, nFlags,;
                                 cMultiByteStr, nSizeMultiByte, ;
                                 @cWideCharStr, nSizeWideChar ;
                               ) ;
            USING STDCALL FROM kernel32.dll

DLLFUNCTION WideCharToMultiByte( nCodePage, nFlags,;
                                 cWideCharStr, nSizeWideChar, ;
                                 @cMultiByteStr, nSizeMultiByte, ;
                                 cDefaultChar, @cUsedDefaultChar ;
                               ) ;
            USING STDCALL FROM kernel32.dll

